/*
Copyright (C) 2014 Graham Jones, www.indriid.com
*/
package stacey.util;

import beast.evolution.tree.Node;
import beast.evolution.tree.TreeInterface;

import java.util.Formatter;
import java.util.Locale;
import java.util.Stack;

/**
 * Created by Graham Jones on 07/11/2014.
 */
public interface UnionNode {


    int getLft();
    int getRgt();
    double getHeight();
    BitUnion getUnion();
    String getID();
    String getExtraText();

    void setLft(int lft);
    void setRgt(int rgt);
    void setAnc(int anc);
    void setHeight(double height);
    void setTipUnionAndID(String id);
    void setBNode(Node bNode);


    public abstract class Abstract implements UnionNode {
        /*
        * Searches subtree rooted at node for most tipward node
        * whose union contains x. If x is known to be a union of one of the nodes,
        * it finds that node, so acts as a map union -> node
        */
        public static int nodeIndexOfUnionInSubtree(UnionNode[] unodes, int nn, BitUnion x) {
            int lft = unodes[nn].getLft();
            int rgt = unodes[nn].getRgt();
            if (lft < 0) {
                return nn;
            }
            if (x.isContainedIn(unodes[lft].getUnion())) {
                return nodeIndexOfUnionInSubtree(unodes, lft, x);
            } else if (x.isContainedIn(unodes[rgt].getUnion())) {
                return nodeIndexOfUnionInSubtree(unodes, rgt, x);
            } else {
                return nn;
            }
        }


        public static void beastTreeToUnionTree(UnionNode[] uNodes, TreeInterface tree) {
            beastSubtreeToUnionSubtree(uNodes, 0, tree.getRoot());
        }



        public static String treeAsText(UnionNode[] uNodes, String extraHeader) {
            int uLength = uNodes[0].getUnion().asText().length();

            String header = "topology                                 idx height  ";
            String uHeader = "union";
            while (uHeader.length() < uLength) {
                uHeader += " ";
            }
            header += uHeader;
            header += extraHeader + "\n";
            String s = "";
            Stack<Integer> x = new Stack<Integer>();
            return header + subtreeAsText(uNodes, uNodes[uNodes.length-1], s, x, 0, "");
        }



        private static String subtreeAsText(UnionNode[] uNodes, UnionNode node, String s, Stack<Integer> x, int depth, String b) {
            Integer[] y = x.toArray(new Integer[x.size()]);
            StringBuffer indent = new StringBuffer();
            for (int i = 0; i < depth; i++) {
                indent.append("  ");
            }
            for (int i = 0; i < y.length; i++) {
                indent.replace(2*y[i], 2*y[i]+1, "|");
            }
            if (b.length() > 0) {
                indent.replace(indent.length()-b.length(), indent.length(), b);
            }
            s += indent;
            s += Abstract.nodeAsText(uNodes, node, indent.length());
            s += System.getProperty("line.separator");
            String subs = "";
            if (node.getLft() >= 0) {
                x.push(depth);
                subs += subtreeAsText(uNodes, uNodes[node.getLft()], "", x, depth+1, "-");
                x.pop();
                subs += subtreeAsText(uNodes, uNodes[node.getRgt()], "", x, depth+1, "`-");
            }
            return s + subs;
        }



        private static String nodeAsText(UnionNode[] uNodes, UnionNode node, int indentlen) {
            StringBuilder s = new StringBuilder();
            Formatter formatter = new Formatter(s, Locale.US);
            if (node.getLft() < 0) {
                formatter.format("%s ", node.getID());
            } else {
                formatter.format("%s ", "+");
            }
            while (s.length() < 40 - indentlen) {
                formatter.format("%s", " ");
            }
            int n = UnionNode.Abstract.nodeIndexOfUnionInSubtree(uNodes, uNodes.length - 1, node.getUnion());
            formatter.format("%3d ", n);
            formatter.format("%s ", Misc.nonnegIn8Chars(node.getHeight()));
            formatter.format("%20s ", node.getUnion().asText());
            formatter.format("%s", node.getExtraText());
            return s.toString();
        }


        private static int beastSubtreeToUnionSubtree(UnionNode[] uNodes, int nextn, Node bNode) {
            if (bNode.isLeaf()) {
                uNodes[nextn].setTipUnionAndID(bNode.getID());
            } else {
                nextn = beastSubtreeToUnionSubtree(uNodes, nextn, bNode.getChild(0));
                int lft = nextn - 1;
                nextn = beastSubtreeToUnionSubtree(uNodes, nextn, bNode.getChild(1));
                int rgt = nextn - 1;
                uNodes[nextn].setLft(lft);
                uNodes[nextn].setRgt(rgt);
                uNodes[lft].setAnc(nextn);
                uNodes[rgt].setAnc(nextn);
                uNodes[nextn].getUnion().union(uNodes[lft].getUnion());
                uNodes[nextn].getUnion().union(uNodes[rgt].getUnion());
            }
            uNodes[nextn].setHeight(bNode.getHeight());
            uNodes[nextn].setBNode(bNode);
            nextn++;
            return nextn;
        }

    }



}
